/* edit-shapebg.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "edit-shapebg.h"
#include "edit-grid.h"
#include "edit-symmetry.h"
#include "xword-state.h"


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  PROP_EDIT_GRID,
  PROP_EDIT_SYMMETRY,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditShapebg
{
  PanelWidget parent_instance;

  PuzzleStack *puzzle_stack;
  GtkWidget *edit_grid;
  GtkWidget *edit_symmetry;
};


G_DEFINE_FINAL_TYPE (EditShapebg, edit_shapebg, PANEL_TYPE_WIDGET);


static void edit_shapebg_init         (EditShapebg           *self);
static void edit_shapebg_class_init   (EditShapebgClass      *klass);
static void edit_shapebg_set_property (GObject               *object,
                                       guint                  prop_id,
                                       const GValue          *value,
                                       GParamSpec            *pspec);
static void edit_shapebg_get_property (GObject               *object,
                                       guint                  prop_id,
                                       GValue                *value,
                                       GParamSpec            *pspec);
static void edit_shapebg_dispose      (GObject               *object);
static void puzzle_changed_cb         (EditShapebg           *self,
                                       gboolean               new_frame,
                                       PuzzleStackChangeType  change_type,
                                       PuzzleStack           *puzzle_stack);

static void
edit_shapebg_class_init (EditShapebgClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_shapebg_set_property;
  object_class->get_property = edit_shapebg_get_property;
  object_class->dispose = edit_shapebg_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-shapebg.ui");

  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_EDIT_GRID] = g_param_spec_object ("edit-grid",
                                                   NULL, NULL,
                                                   EDIT_TYPE_GRID,
                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  obj_props[PROP_EDIT_SYMMETRY] = g_param_spec_object ("edit-symmetry",
                                                       NULL, NULL,
                                                       EDIT_TYPE_SYMMETRY,
                                                       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_shapebg_init (EditShapebg *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));
}

static void
edit_shapebg_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
  EditShapebg *self = EDIT_SHAPEBG (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_GRID:
      self->edit_grid = g_value_dup_object (value);
      break;
    case PROP_EDIT_SYMMETRY:
      self->edit_symmetry = g_value_dup_object (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_shapebg_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
{
  EditShapebg *self = EDIT_SHAPEBG (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_GRID:
      g_value_set_object (value, self->edit_grid);
      break;
    case PROP_EDIT_SYMMETRY:
      g_value_set_object (value, self->edit_symmetry);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_shapebg_dispose (GObject *object)
{
  EditShapebg *self;
  GtkWidget *child;

  self = EDIT_SHAPEBG (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  /* We have refs to these */
  g_clear_object (&self->puzzle_stack);
  g_clear_object (&self->edit_grid);
  g_clear_object (&self->edit_symmetry);

  G_OBJECT_CLASS (edit_shapebg_parent_class)->dispose (object);
}


static void
edit_shapebg_update_state (EditShapebg *self)
{
}

/* Public methods */

static void
puzzle_changed_cb (EditShapebg              *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  if (change_type == STACK_CHANGE_PUZZLE || change_type == STACK_CHANGE_GRID)
    edit_shapebg_update_state (self);
}
