/* edit-word-list.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "crosswords-enums.h"
#include "edit-grid.h"
#include "edit-word-list.h"
#include "play-grid.h"
#include "word-list.h"
#include "word-list-model.h"


enum
{
  WORD_ACTIVATED,
  N_SIGNALS
};

static guint obj_signals[N_SIGNALS] = { 0 };


struct _EditWordList
{
  PanelWidget parent_instance;

  WordListModel *across_list_model;
  WordListModel *down_list_model;

  GtkWidget *across_word_list_view;
  GtkWidget *down_word_list_view;
};


G_DEFINE_TYPE (EditWordList, edit_word_list, PANEL_TYPE_WIDGET);


static void edit_word_list_init          (EditWordList          *self);
static void edit_word_list_class_init    (EditWordListClass     *klass);
static void across_word_activate_cb      (EditWordList          *self,
                                          guint                  position);
static void down_word_activate_cb        (EditWordList          *self,
                                          guint                  position);


static void
setup_listitem_cb (GtkListItemFactory *factory,
                   GtkListItem        *list_item)
{
  GtkWidget *label;

  label = gtk_label_new (NULL);
  gtk_label_set_xalign (GTK_LABEL (label), 0.0);
  gtk_list_item_set_child (list_item, label);
}

static void
row_changed (WordListModelRow *row,
             GtkWidget        *label)
{
  g_autofree gchar *text = NULL;

  text = g_strdup_printf ("%s — (%d)",
                          word_list_model_row_get_word (row),
                          word_list_model_row_get_priority (row));
  gtk_label_set_text (GTK_LABEL (label), text);
}

static void
bind_listitem_cb (GtkListItemFactory *factory,
                  GtkListItem        *list_item)
{
  GtkWidget *label;
  WordListModelRow *row;

  label = gtk_list_item_get_child (list_item);
  row = (WordListModelRow *) gtk_list_item_get_item (list_item);

  g_signal_connect (row, "changed", G_CALLBACK (row_changed), label);
  row_changed (row, label);
}

static void
unbind_listitem_cb (GtkListItemFactory *factory,
                    GtkListItem        *list_item)
{
  GtkWidget *label;
  WordListModelRow *row;

  label = gtk_list_item_get_child (list_item);
  row = (WordListModelRow *) gtk_list_item_get_item (list_item);

  g_signal_handlers_disconnect_by_func (row, G_CALLBACK (row_changed), label);
}

static void
edit_word_list_init (EditWordList *self)
{
  g_autoptr (GtkListItemFactory) across_factory = NULL;
  g_autoptr (GtkListItemFactory) down_factory = NULL;

  self->across_list_model = word_list_model_new ();
  self->down_list_model = word_list_model_new ();

  gtk_widget_init_template (GTK_WIDGET (self));

  across_factory = gtk_signal_list_item_factory_new ();
  g_signal_connect (across_factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (across_factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);
  g_signal_connect (across_factory, "unbind", G_CALLBACK (unbind_listitem_cb), NULL);

  down_factory = gtk_signal_list_item_factory_new ();
  g_signal_connect (down_factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (down_factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);
  g_signal_connect (down_factory, "unbind", G_CALLBACK (unbind_listitem_cb), NULL);

  g_object_set (self->across_word_list_view,
                "model", gtk_no_selection_new (G_LIST_MODEL (self->across_list_model)),
                "factory", across_factory,
                NULL);
  g_object_set (self->down_word_list_view,
                "model", gtk_no_selection_new (G_LIST_MODEL (self->down_list_model)),
                "factory", down_factory,
                NULL);
}

static void
edit_word_list_class_init (EditWordListClass *klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-word-list.ui");

  gtk_widget_class_bind_template_callback (widget_class, across_word_activate_cb);
  gtk_widget_class_bind_template_callback (widget_class, down_word_activate_cb);

  gtk_widget_class_bind_template_child (widget_class, EditWordList, across_word_list_view);
  gtk_widget_class_bind_template_child (widget_class, EditWordList, down_word_list_view);

  obj_signals [WORD_ACTIVATED] =
    g_signal_new ("word-activated",
                  EDIT_TYPE_WORD_LIST,
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 2,
                  G_TYPE_STRING,
                  I_TYPE_PUZ_CLUE_DIRECTION);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
}

static void
word_activate (EditWordList      *self,
               guint              position,
               IPuzClueDirection  direction)
{
  GListModel *model;
  WordListModelRow *row;
  const gchar *word;

  if (direction == IPUZ_CLUE_DIRECTION_ACROSS)
    model = G_LIST_MODEL (self->across_list_model);
  else
    model = G_LIST_MODEL (self->down_list_model);

  row = g_list_model_get_item (model, position);
  word = word_list_model_row_get_word (row);
  g_assert (word);

  g_signal_emit (self, obj_signals [WORD_ACTIVATED], 0, word, direction);
}

static void
across_word_activate_cb (EditWordList *self,
                         guint      position)
{
  word_activate (self, position, IPUZ_CLUE_DIRECTION_ACROSS);
}

static void
down_word_activate_cb (EditWordList *self,
                       guint      position)
{
  word_activate (self, position, IPUZ_CLUE_DIRECTION_DOWN);
}

void
edit_word_list_update (EditWordList *self,
                       XwordState   *edit_state)
{
  IPuzClue *across_clue;
  IPuzClue *down_clue;
  g_autofree gchar *across_clue_string = NULL;
  g_autofree gchar *down_clue_string = NULL;

  g_return_if_fail (EDIT_IS_WORD_LIST (self));

  if (edit_state->clue.direction == IPUZ_CLUE_DIRECTION_NONE)
    {
      word_list_model_set_filter (self->across_list_model, NULL, WORD_LIST_MATCH);
      word_list_model_set_filter (self->down_list_model, NULL, WORD_LIST_MATCH);
      return;
    }

  across_clue = ipuz_crossword_find_clue_by_coord (edit_state->xword,
                                                   IPUZ_CLUE_DIRECTION_ACROSS,
                                                   edit_state->cursor);
  down_clue = ipuz_crossword_find_clue_by_coord (edit_state->xword,
                                                 IPUZ_CLUE_DIRECTION_DOWN,
                                                 edit_state->cursor);

  if (across_clue)
    {
      IPuzClueId clue_id = ipuz_crossword_get_clue_id (edit_state->xword, across_clue);
      across_clue_string = ipuz_crossword_get_clue_string_by_id (edit_state->xword, clue_id);
    }

  if (down_clue)
    {
      IPuzClueId clue_id = ipuz_crossword_get_clue_id (edit_state->xword, down_clue);
      down_clue_string = ipuz_crossword_get_clue_string_by_id (edit_state->xword, clue_id);
    }

  word_list_model_set_filter (self->across_list_model, across_clue_string, WORD_LIST_MATCH);
  word_list_model_set_filter (self->down_list_model, down_clue_string, WORD_LIST_MATCH);
}
